@@ -14,18 +14,24 @@
 """Python BTree implementation
 """
 
-from struct import Struct
-from struct import error as struct_error
-from operator import index
+
 
 from persistent import Persistent
 
 from .Interfaces import BTreesConflictError
 from ._compat import PY3
 from ._compat import compare
-from ._compat import int_types
 from ._compat import xrange
 
+# XXX: Fix these. These ignores are temporary to
+# reduce the noise and help find specific issues during
+# refactoring
+# pylint:disable=too-many-lines,fixme,protected-access
+# pylint:disable=attribute-defined-outside-init,redefined-builtin,no-else-return
+# pylint:disable=redefined-outer-name,bad-continuation,unused-variable
+# pylint:disable=too-many-branches,too-many-statements,arguments-differ,assigning-non-slot
+# pylint:disable=superfluous-parens,inconsistent-return-statements,unidiomatic-typecheck
+# pylint:disable=deprecated-method,consider-using-enumerate
 
 _marker = object()
 
@@ -33,6 +39,11 @@ _marker = object()
 class _Base(Persistent):
 
     __slots__ = ()
+    # This is used to allocate storage for the keys.
+    # It's probably here so that we could, for example, use
+    # an ``array.array`` for native types. But nothing actually does
+    # that, everything is stored boxed.
+    # TODO: Figure out why not.
     _key_type = list
 
     def __init__(self, items=None):
@@ -79,7 +90,30 @@ class _Base(Persistent):
 
         _BTree_reduce_up_bound = _BTree_reduce_as
 
-class _BucketBase(_Base):
+class _ArithmeticMixin(object):
+    def __sub__(self, other):
+        return difference(self.__class__, self, other)
+
+    def __rsub__(self, other):
+        return difference(self._set_type, type(self)(other), self)
+
+    def __or__(self, other):
+        return union(self._set_type, self, other)
+
+    __ror__ = __or__
+
+    def __and__(self, other):
+        return intersection(self._set_type, self, other)
+
+    __rand__ = __and__
+
+    def __xor__(self, other):
+        return (self - other) | (other - self)
+
+    __rxor__ = __xor__
+
+
+class _BucketBase(_ArithmeticMixin, _Base):
 
     __slots__ = ('_keys',
                  '_next',
@@ -190,7 +224,13 @@ class _BucketBase(_Base):
         return iter(self._keys)
 
     def __contains__(self, key):
-        return (self._search(self._to_key(key)) >= 0)
+        try:
+            tree_key = self._to_key(key)
+        except TypeError:
+            # Can't convert the key, so can't possibly be in the tree
+            return False
+
+        return (self._search(tree_key) >= 0)
 
     has_key = __contains__
 
@@ -201,6 +241,7 @@ class _BucketBase(_Base):
         name = name[:-2] if name.endswith("Py") else name
         return "%s.%s(%r)" % (mod, name, items)
 
+
 class _SetIteration(object):
 
     __slots__ = ('to_iterate',
@@ -252,27 +293,26 @@ class _SetIteration(object):
 
         return self
 
-_object_lt = getattr(object, '__lt__', _marker)
+class _MutableMappingMixin(object):
+    # Methods defined in collections.abc.MutableMapping that
+    # Bucket and Tree should both implement and can implement
+    # the same. We don't want to extend that class though,
+    # as the C version cannot.
+    def popitem(self):
+        """
+        D.popitem() -> (k, v), remove and return some (key, value) pair
+        as a 2-tuple; but raise KeyError if D is empty.
+        """
+        try:
+            key = next(iter(self))
+        except StopIteration:
+            raise KeyError
+        value = self[key]
+        del self[key]
+        return key, value
 
-def _no_default_comparison(key):
-    # Enforce test that key has non-default comparison.
-    if key is None:
-        return
-    if type(key) is object:
-        raise TypeError("Can't use object() as keys")
-    lt = getattr(key, '__lt__', None)
-    if lt is not None:
-        # CPython 3.x follows PEP 252, defining '__objclass__'
-        if getattr(lt, '__objclass__', None) is object:
-            lt = None  # pragma: no cover Py3k
-        # PyPy3 doesn't follow PEP 252, but defines '__func__'
-        elif getattr(lt, '__func__', None) is _object_lt:
-            lt = None  # pragma: no cover PyPy3
-    if (lt is None and
-        getattr(key, '__cmp__', None) is None):
-        raise TypeError("Object has default comparison")
 
-class Bucket(_BucketBase):
+class Bucket(_MutableMappingMixin, _BucketBase):
 
     __slots__ = ()
     _value_type = list
@@ -308,7 +348,6 @@ class Bucket(_BucketBase):
             raise TypeError('items must be a sequence of 2-tuples')
 
     def __setitem__(self, key, value):
-        _no_default_comparison(key)
         self._set(self._to_key(key), self._to_value(value))
 
     def __delitem__(self, key):
@@ -319,13 +358,23 @@ class Bucket(_BucketBase):
         self._values = self._value_type()
 
     def get(self, key, default=None):
-        index = self._search(self._to_key(key))
+        try:
+            key = self._to_key(key)
+        except TypeError:
+            # Can't convert, cannot possibly be present.
+            return default
+        index = self._search(key)
         if index < 0:
             return default
         return self._values[index]
 
     def __getitem__(self, key):
-        index = self._search(self._to_key(key))
+        try:
+            tree_key = self._to_key(key)
+        except TypeError:
+            # Can't convert, so cannot possibly be present.
+            raise KeyError(key)
+        index = self._search(tree_key)
         if index < 0:
             raise KeyError(key)
         return self._values[index]
@@ -565,7 +614,70 @@ class Bucket(_BucketBase):
     def __repr__(self):
         return self._repr_helper(self.items())
 
-class Set(_BucketBase):
+class _MutableSetMixin(object):
+    # Like _MutableMappingMixin, but for sets.
+    def isdisjoint(self, other):
+        """
+        Return True if two sets have a null intersection.
+        """
+        for value in other:
+            if value in self:
+                return False
+        return True
+
+    def discard(self, key):
+        """
+        Remove an element from the set if it is a member.
+
+        If not, do nothing and raise no exception.
+        """
+        # Written this way to avoid catching and accidentally
+        # ignoring POSKeyError.
+        if key in self:
+            self.remove(key)
+
+    def pop(self):
+        """Return the popped value.  Raise KeyError if empty."""
+        # Get our iter first to avoid catching and accidentally
+        # ignoring POSKeyError
+        it = iter(self)
+        try:
+            value = next(it)
+        except StopIteration:
+            raise KeyError
+        self.discard(value)
+        return value
+
+    def __ior__(self, it):
+        self.update(it)
+        return self
+
+    def __iand__(self, it):
+        for value in (self - it):
+            self.discard(value)
+        return self
+
+    def __isub__(self, it):
+        if it is self:
+            self.clear()
+        else:
+            for value in it:
+                self.discard(value)
+        return self
+
+    def __ixor__(self, it):
+        if it is self:
+            self.clear()
+        else:
+            for value in it:
+                if value in self:
+                    self.discard(value)
+                else:
+                    self.add(value)
+        return self
+
+
+class Set(_MutableSetMixin, _BucketBase):
 
     __slots__ = ()
 
@@ -772,7 +884,7 @@ class _TreeItem(object):
         self.child = child
 
 
-class _Tree(_Base):
+class _Tree(_ArithmeticMixin, _Base):
 
     __slots__ = ('_data',
                  '_firstbucket',
@@ -809,7 +921,6 @@ class _Tree(_Base):
             set(*i)
 
     def __setitem__(self, key, value):
-        _no_default_comparison(key)
         self._set(self._to_key(key), self._to_value(value))
 
     def __delitem__(self, key):
@@ -990,7 +1101,7 @@ class _Tree(_Base):
         if isinstance(first.child, type(self)):
             next._firstbucket = first.child._firstbucket
         else:
-            next._firstbucket = first.child;
+            next._firstbucket = first.child
         return next
 
     def _del(self, key):
@@ -1034,6 +1145,7 @@ class _Tree(_Base):
                     self._firstbucket = child._next
                     removed_first_bucket = True
             del data[index]
+            self._p_changed = True
 
         return removed_first_bucket, value
 
@@ -1054,13 +1166,12 @@ class _Tree(_Base):
             ):
             return ((data[0].child.__getstate__(), ), )
 
-        sdata = []
+
+        data = iter(data)
+        sdata = [next(data).child]
         for item in data:
-            if sdata:
-                sdata.append(item.key)
-                sdata.append(item.child)
-            else:
-                sdata.append(item.child)
+            sdata.append(item.key)
+            sdata.append(item.child)
 
         return tuple(sdata), self._firstbucket
 
@@ -1080,6 +1191,16 @@ class _Tree(_Base):
         data, self._firstbucket = state
         data = list(reversed(data))
 
+        # verify children are either tree or bucket nodes.
+        # NOTE for tree-kind node type is compared as "is", not as
+        # "isinstance", to match C version.
+        for child in data[::2]:
+            if not ((type(child) is type(self)) or
+                    isinstance(child, self._bucket_type)):
+                raise TypeError("tree child %s is neither %s nor %s" %
+                                (_tp_name(type(child)), _tp_name(type(self)),
+                                 _tp_name(self._bucket_type)))
+
         self._data.append(_TreeItem(None, data.pop()))
         while data:
             key = data.pop()
@@ -1112,7 +1233,7 @@ class _Tree(_Base):
                     "BTree has firstbucket different than "
                     "its first child's firstbucket")
             for i in range(len(data)-1):
-               data[i].child._check(data[i+1].child._firstbucket)
+                data[i].child._check(data[i+1].child._firstbucket)
             data[-1].child._check(nextbucket)
         elif child_class is self._bucket_type:
             assert_(self._firstbucket is data[0].child,
@@ -1138,6 +1259,7 @@ class _Tree(_Base):
         r = r.replace('Py', '')
         return r
 
+
 def _get_simple_btree_bucket_state(state):
     if state is None:
         return state
@@ -1236,7 +1358,7 @@ class _TreeIterator(object):
         )
 
 
-class Tree(_Tree):
+class Tree(_MutableMappingMixin, _Tree):
 
     __slots__ = ()
 
@@ -1276,7 +1398,7 @@ class Tree(_Tree):
         return bool(self._set(key, value, True)[0])
 
 
-class TreeSet(_Tree):
+class TreeSet(_MutableSetMixin, _Tree):
 
     __slots__ = ()
 
@@ -1298,13 +1420,23 @@ class TreeSet(_Tree):
 
 class set_operation(object):
 
-    __slots__ = ('func',
-                 'set_type',
-                )
+    __slots__ = (
+        'func',
+        'set_type',
+        '__name__',
+        '_module',
+    )
 
     def __init__(self, func, set_type):
         self.func = func
         self.set_type = set_type
+        self.__name__ = func.__name__
+        self._module = func.__module__
+
+    __module__ = property(
+        lambda self: self._module,
+        lambda self, nv: setattr(self, '_module', nv)
+    )
 
     def __call__(self, *a, **k):
         return self.func(self.set_type, *a, **k)
@@ -1346,7 +1478,7 @@ def union(set_type, o1, o2):
         return o1
     i1 = _SetIteration(o1, False, 0)
     i2 = _SetIteration(o2, False, 0)
-    result = o1._set_type()
+    result = set_type()
     def copy(i):
         result._keys.append(i.key)
     while i1.active and i2.active:
@@ -1376,7 +1508,7 @@ def intersection(set_type, o1, o2):
         return o1
     i1 = _SetIteration(o1, False, 0)
     i2 = _SetIteration(o2, False, 0)
-    result = o1._set_type()
+    result = set_type()
     def copy(i):
         result._keys.append(i.key)
     while i1.active and i2.active:
@@ -1493,56 +1625,6 @@ def multiunion(set_type, seqs):
         result.update(s)
     return result
 
-def to_ob(self, v):
-    return v
-
-def _packer_unpacker(struct_format):
-    s = Struct(struct_format)
-    return s.pack, s.unpack
-
-int_pack, int_unpack = _packer_unpacker('i')
-
-def to_int(self, v):
-    try:
-        int_pack(index(v))
-    except (struct_error, TypeError):
-        raise TypeError('32-bit integer expected')
-
-    return int(v)
-
-float_pack = _packer_unpacker('f')[0]
-
-def to_float(self, v):
-    try:
-        float_pack(v)
-    except struct_error:
-        raise TypeError('float expected')
-    return float(v)
-
-
-long_pack, long_unpack = _packer_unpacker('q')
-
-def to_long(self, v):
-    try:
-        long_pack(index(v))
-    except (struct_error, TypeError):
-        if isinstance(v, int_types):
-            raise ValueError("Value out of range", v)
-        raise TypeError('64-bit integer expected')
-
-    return int(v)
-
-def to_bytes(l):
-    def to(self, v):
-        if not (isinstance(v, bytes) and len(v) == l):
-            raise TypeError("%s-byte array expected" % l)
-        return v
-    return to
-
-tos = dict(I=to_int, L=to_long, F=to_float, O=to_ob)
-
-MERGE_DEFAULT_int = 1
-MERGE_DEFAULT_float = 1.0
 
 def MERGE(self, value1, weight1, value2, weight2):
     return (value1 * weight1) + (value2 * weight2)
@@ -1593,3 +1675,9 @@ def _fix_pickle(mod_dict, mod_name):
             # Python pickle match the C version by default
             py_type.__name__ = raw_name
             py_type.__qualname__ = raw_name # Py 3.3+
+
+
+# tp_name returns full name of a type in the same way as how it is provided by
+# typ->tp_name in C.
+def _tp_name(typ):
+    return '.'.join([typ.__module__, typ.__name__])
